#!/usr/bin/env python3

import sys
import os
from pathlib import Path

def find_mkl_config(prefix_path: str) -> "Path | None":
    for root, dirs, files in os.walk(prefix_path):
        for file in files:
            if file == "MKLConfig.cmake":
                return Path(root) / file
    return None


def run():
    for prefix_path in sys.argv + [sys.prefix]:
        mkl_config = find_mkl_config(prefix_path)
        if mkl_config:
            print(mkl_config)
            sys.exit(0)
    sys.exit(1)

if __name__ == '__main__':
    run()



import logging
import os
import re
import subprocess
import sysconfig
from pathlib import Path
from typing import Iterable, Optional, List, Union


def preprocess(file: Path, include_directories: Optional[List[str]] = None) -> Optional[List[str]]:
    cxx = os.environ.get("CXX")
    if cxx:
        # GCC, and older versions of clang may fail if it cannot find references to the #include directives
        # Additionally, the result will not be annotated with -E -dI
        # Clang does not require it, though
        if include_directories is None:
            additional_include_directories = []
        else:
            additional_include_directories = [Path(directory) for directory in include_directories]
        # MSVC uses different arguments and notation than gcc / llvm
        if re.search(r"^[a-z]:[\\/].*[\\/]Microsoft Visual Studio[\\/].*[\\/]cl.exe", cxx, re.IGNORECASE):
            preprocessor_args = ("/P", "/showIncludes")
            include_arg = "/I"
        else:
            preprocessor_args = ("-E", "-dI")
            include_arg = "-I"
        cxx_flags = os.environ.get("CXXFLAGS") or ""
        res = subprocess.run(" ".join([
            f'"{cxx}"',
            *cxx_flags.split(" "),
            *preprocessor_args,
            *[f'{include_arg}"{include}"' for include in additional_include_directories],
            str(file),
        ]),
            check=False,
            capture_output=True,
            shell=True,
        )
        return [
            line[0].decode("utf-8").strip()
            for line in re.findall(b'^((# *[0-9]+ +|Note: including file: +|.*-E -dI).*$)', res.stdout + res.stderr, re.MULTILINE)
        ]


def is_relative_to(path: Path, other: Union[Path, str]) -> bool:
    try:
        return path.is_relative_to(other)
    except AttributeError:
        # Backport for Python < 3.9
        try:
            path.relative_to(other)
            return True
        except ValueError:
            return False


def collect_sources(
        from_source_files: Iterable[str],
        ignore: Optional[List[str]] = None,
        source_dir: str = 'src',
        use_preprocessor: bool = True,
        include_directories: Optional[List[str]] = None,
) -> List[str]:
    if ignore is None:
        ignore = []
    files = []

    src = Path(source_dir)
    root = Path('.').absolute()
    _files_to_be_inspected = [Path(file) for file in from_source_files]
    files_to_be_inspected = []
    for file in _files_to_be_inspected:  # type: Path
        if file.is_dir():
            for child in file.glob('*'):
                if child.is_file():
                    files_to_be_inspected.append(child)
        else:
            files_to_be_inspected.append(file)

    header_regex = re.compile(r'\.h(pp)?$')
    source_regex = re.compile(r'\.c(pp)?$')

    def add_if_necessary(file):
        if any(str(file).startswith(ignore_folder) for ignore_folder in ignore):
            return
        if file not in files and file not in files_to_be_inspected:

            def _add_file(repl: str, file_regex):
                target_file = file.parent / file_regex.sub(repl, file.name)
                if target_file.exists():
                    files_to_be_inspected.append(target_file)

            files_to_be_inspected.append(file)
            # TODO: Deal with dSFT.cpp has a .h file
            if file.suffix in ['.hpp', '.h']:
                _add_file(r'.c\1', header_regex)
                _add_file(r'.cpp', header_regex)
            if file.suffix in ['.cpp', '.c']:
                _add_file(r'.h\1', source_regex)

    _name_group = r'(?P<name>([a-z]:[\\/])?[\w\\/._-]+)'
    _name_pattern = r'[<"]' + _name_group + '[>"]'
    include_pattern = re.compile(r'^ *# *include *' + _name_pattern, re.IGNORECASE)
    define_pattern = re.compile(r'^ *# *define \w+ *(\\\n)? *' + _name_pattern, re.IGNORECASE | re.MULTILINE)
    completely_preprocessed_pattern = re.compile(r'^ *# *\d+ *' + _name_pattern + r'( \d+)*', re.IGNORECASE)
    msvc_include_pattern = re.compile(r'^Note: including file: +' + _name_group, re.IGNORECASE)
    while len(files_to_be_inspected) > 0:

        file = files_to_be_inspected.pop().resolve()
        try:
            file = file.relative_to(root)
        except ValueError:
            # Skip files from the system
            continue
        if use_preprocessor and is_relative_to(file, sysconfig.get_path("include")):
            # We are not interested in these files, and some of them should not be used directly
            # which we'll be doing here
            continue
        name = str(file)
        if name in files:
            continue

        with file.open(encoding='utf8') as f:
            if use_preprocessor:
                source_lines = preprocess(file, include_directories)
            if not use_preprocessor or source_lines is None:
                source_lines = f.readlines()
        try:
            previous_line = ''
            for line in source_lines:
                match = include_pattern.search(line)
                if not match:
                    match = define_pattern.search(line)
                if not match:
                    if not previous_line.endswith('\n'):
                        previous_line += '\n'
                    match = define_pattern.search(previous_line + line)
                if not match:
                    match = completely_preprocessed_pattern.search(line)
                if not match:
                    match = msvc_include_pattern.search(line)
                if match:
                    item = Path(match.group('name'))
                    if (file.parent / item).is_file():
                        add_if_necessary(file.parent / item)
                    elif item.is_file():
                        add_if_necessary(item)
                    elif (src / item).is_file():
                        add_if_necessary(src / item)
                    else:
                        pass
                previous_line = line
        except UnicodeDecodeError:
            logging.info(f"'{name}' could not be opened / decoded as a text file. It's been ignored")

        files.append(name)

    # CMake expects UNIX style paths even on Windows
    files = [str(file).replace('\\', '/') for file in files]
    return files



import subprocess
import sys
try:
    import tomllib as toml
    mode = 'rb'
except ImportError:
    import toml
    mode = 'r'


with open("pyproject.toml", mode) as f:
    pyproject = toml.load(f)

# Install required dependencies
requirements = pyproject['build-system']['requires']

with open('requirements.txt', 'w') as f:
    f.write("\n".join(requirements))

subprocess.run(
    " ".join([
        sys.executable,
        "-m",
        "pip",
        "install",
        "-r",
        "requirements.txt",
    ]),
    shell=True,
    check=True,
)



try:
    import scipy
except ImportError:
    raise ImportError("Scipy is not installed, and is required to use these utilities")

from gaussianfft.util.empirical_variogram import EmpiricalVariogram
from gaussianfft.util.padding_analyzer import PaddingAnalyzerDefaults, PaddingAnalyzer

__all__ = ['EmpiricalVariogram', 'PaddingAnalyzer', 'PaddingAnalyzerDefaults']



from time import perf_counter

import numpy as np
import gaussianfft

from gaussianfft.util import EmpiricalVariogram


class PaddingAnalyzerDefaults:
    DRF = 1.0
    L = 1000
    n = 100
    seed = 12313
    nmax = 600
    ndims = 1


class PaddingAnalyzer:
    """
    Class for analyzing the relationship between range and padding and how it affects the estimation of the true
    variogram.
    """
    def __init__(self, desired_range, desired_padding_f, L=None, n=None, drf=None, nmax=None, vtype='gaussian',
                 seed=None, ndims=None):
        # Get default values
        n = n or PaddingAnalyzerDefaults.n
        drf = drf or PaddingAnalyzerDefaults.DRF
        L = L or PaddingAnalyzerDefaults.L
        seed = seed or PaddingAnalyzerDefaults.seed
        nmax = nmax or PaddingAnalyzerDefaults.nmax
        ndims = ndims or PaddingAnalyzerDefaults.ndims

        # Constants that potentially can be moved to input
        self._reference_strategy = 'origo'
        self._convergence_step = 10
        # Setup
        gaussianfft.seed(seed)
        d = L / n
        dr = drf * d

        # Convert from aprx pl to exact
        output_padding = []  # Actual padding (in number of gridcells)
        for p in desired_padding_f:
            output_padding.append(int((p * L) / d))

        # Prepare results
        v, ev = self.create_variograms(desired_range[0], output_padding[0], ndims, d, n, L, vtype)
        refs = ev.pick_reference_points(self._reference_strategy)
        mid, _, _, _, convrg = ev.estimate_variogram(nmax, dr, refs, analyze_convergence=self._convergence_step)
        # -----

        self.deltas = np.zeros((len(output_padding), len(desired_range), convrg.deltas.shape[1]))
        t0 = perf_counter()
        for i, r in enumerate(desired_range):
            print("*** {}/{} ***".format(i, len(desired_range)))
            print("  Total time spent: {:.2f}".format(perf_counter() - t0))
            for j, p in enumerate(output_padding):
                v, ev = self.create_variograms(r, p, ndims, d, n, L, vtype)
                refs = ev.pick_reference_points('origo')
                mid, _, _, _, convrg = ev.estimate_variogram(nmax, dr, refs, analyze_convergence=self._convergence_step)
                self.deltas[j, i, :] = convrg.deltas[-1]

        self.range = desired_range
        self.padding_f = np.array(output_padding) / n
        self.grid = np.linspace(0, self.deltas.shape[2] * d, self.deltas.shape[2])

    def create_variograms(self, rrange, padding, ndims, d, n, L, vtype):
        if ndims == 1:
            v = gaussianfft.variogram(vtype, rrange * L)
            ev = EmpiricalVariogram(v, n, d, 1, 0, 1, 0, padding, padding, padding)
        elif ndims == 2:
            v = gaussianfft.variogram(vtype, rrange * L, rrange * L)
            ev = EmpiricalVariogram(v, n, d, n, d, 1, 0, padding, padding, padding)
        else:
            raise NotImplementedError("Padding analyzer is not implemented for {} dimensions.".format(ndims))
        return v, ev



import random
from time import perf_counter

import gaussianfft
import numpy as np
from scipy.spatial.distance import cdist
from scipy.stats import binned_statistic

from gaussianfft.util.empirical_variogram2 import EmpiricalVariogram2


class EmpiricalVariogram:
    def __init__(self, v: gaussianfft.Variogram, nx, dx, ny, dy, nz, dz, px, py, pz):
        self.nx, self.ny, self.nz = nx, ny, nz
        self.dx, self.dy, self.dz = dx, dy, dz
        self.px, self.py, self.pz = px, py, pz
        self.v = v
        self.points, self.indexes = EmpiricalVariogram.find_grid_points(self.nx, self.dx,
                                                                        self.ny, self.dy,
                                                                        self.nz, self.dz)
        self.tuple_indexes = [tuple(i) for i in self.indexes]

    def estimate_variogram(self, nmax, resolution, reference_points, analyze_convergence=0):
        """
        Main function of this class. Used to estimate the variogram.

        :param nmax:                The number of simulations
        :param resolution:          Resolution of the output variogram
        :param reference_points:    List of length 3 tuples with each tuple representing a grid point from where the
                                    variogram is estimated.
        :param analyze_convergence: Sample deviation from the true variogram at an interval determined by this value.
                                    If <= 0, then this is ignored, which will reduce memory consumption.
        :return:
            [0] distance values, shape (N, )
            [1] estimated variogram at each distance value, shape (N, )
        """
        midpoints = self.find_midpoints(resolution)
        accumulated_sum = np.zeros_like(midpoints)
        n_samples = np.zeros_like(midpoints)
        tdata = {'sim': [], 'est': []}
        ca = ConvergenceAnalyzer(midpoints, self.true_variogram(resolution)[1])
        for i in range(nmax):
            t0 = perf_counter()
            s = self.simulate()
            t1 = perf_counter() - t0
            tdata['sim'].append(t1)
            t2 = perf_counter()
            # Find the mean for all reference points
            local_acc_mean = np.zeros_like(accumulated_sum)
            local_count = np.zeros_like(local_acc_mean)
            for ref in reference_points:
                means, _ = self.binned_variogram_statistics(resolution, s, ref)
                means = 1 - means / 2.0  # This will ensure an unbiased estimator for the variogram
                len_diff = len(accumulated_sum) - len(means)
                local_acc_mean += np.pad(means, (0, len_diff), 'constant')
                local_count += np.pad(np.ones_like(means), (0, len_diff), 'constant')
            # May divide on zero here, but that is okay. Interpreting nan as missing data will be the correct
            # interpretation
            local_acc_mean /= local_count
            # Add the means to the total accumulated
            accumulated_sum += local_acc_mean
            n_samples += local_count
            t3 = perf_counter() - t2
            tdata['est'].append(t3)
            if analyze_convergence > 0 and not i % analyze_convergence:
                ca.feed(accumulated_sum/(i+1))
        accumulated_mean = accumulated_sum / nmax
        return midpoints, accumulated_mean, n_samples, tdata, ca

    def true_variogram(self, resolution):
        """
        Useful for comparison with the true variogram
        :param resolution:
        :return:
        """
        midpoints = self.find_midpoints(resolution)
        true_variogram = []
        for m in midpoints:
            # This will only work for a variogram with equal ranges in each dimension. Should otherwise use
            # self.v.corr(m1, m2[, m3]) instead
            true_variogram.append(self.v.corr(m))
        return midpoints, np.array(true_variogram)

    def pick_reference_points(self, strategy, n=0, offset=0, seed=None):
        """
        Utility function for finding reference points based on a high level strategy instead of low level control
        of reference points.
        :param strategy: The strategy to use. See below for options
        :param n:        The (approximate) number of points. Valid for 'random' and 'regular' strategies
        :param offset:   First tupled index to use. Valid for 'regular' stategy
        :param seed:     Sets the random seed. Valid for 'random' strategy
        :return:
        """
        if seed is not None:
            random.seed(seed)
        if strategy == 'all':
            reference_points = self.tuple_indexes
        elif strategy == 'origo':
            reference_points = [(0, 0, 0)]
        elif strategy == 'center':
            reference_points = [(int(self.nx/2), int(self.ny/2), int(self.nz/2))]
        elif strategy == 'random':
            reference_points = random.sample(self.tuple_indexes, n)
        elif strategy == 'regular':
            reference_points = self.tuple_indexes[offset::int(len(self.tuple_indexes)/n)]
        else:
            raise ValueError("Invalid strategy '{}'".format(strategy))
        return reference_points

    """********************************************************"""
    """ Primarily utility functions used by estimate_variogram """
    """********************************************************"""

    def calculate_deltas(self, grid, ref=None):
        ref = ref or (0, 0, 0)
        if self.ny == 1:
            grid = grid.reshape((self.nx, 1, 1))
        elif self.nz == 1:
            grid = grid.reshape((self.nx, self.ny, 1))

        p_idx = self.tuple_indexes.index(ref)
        dists = cdist(self.points[p_idx].reshape((1, 3)), self.points)
        delta = grid - grid[ref]
        delta_flat = delta[self.indexes[:, 0], self.indexes[:, 1], self.indexes[:, 2]]

        return dists, delta_flat

    def binned_variogram_statistics(self, resolution, grid, ref=None):
        dists, deltas = self.calculate_deltas(grid, ref)
        dists = dists.flatten()
        deltas2 = np.square(deltas)
        nbins = int(np.max(dists)/resolution)
        mean_stats, edges, _ = binned_statistic(dists, deltas2, statistic='mean', bins=nbins)
        return mean_stats, edges

    def find_midpoints(self, resolution):
        s = self.simulate()
        means, edges = self.binned_variogram_statistics(resolution, s)
        return edges[:-1] + np.diff(edges)[0]

    def simulate(self):
        s = gaussianfft.advanced.simulate(self.v,
                                    self.nx, self.dx,
                                    self.ny, self.dy,
                                    self.nz, self.dz,
                                    padx=self.px, pady=self.py, padz=self.pz)
        s = np.array(s).reshape((self.nx, self.ny, self.nz), order='F')
        return s

    def dummy_sim(self):
        # Simulate dummy data with no variance, but exact variogram generation.
        # Can be used to verify that estimate_variogram works as intended, e.g.
        # by monkey patching the simulate function before estimation is done:
        #   e = EmpiricalVariogram(...)
        #   e.simulate = e.dummy_sim
        #   result = e.estimate_variogram(...)
        ref = (0, 0, 0)
        p_idx = self.tuple_indexes.index(ref)
        dists = cdist(self.points[p_idx].reshape((1, 3)), self.points)
        dists = dists.flatten()
        s_values = [np.sqrt((1-self.v.corr(d)) * 2) for d in dists]
        s = np.zeros((self.nx, self.ny, self.nz))
        s[self.indexes[:, 0], self.indexes[:, 1], self.indexes[:, 2]] = s_values
        return s

    @staticmethod
    def find_grid_points(nx, dx, ny=1, dy=0.0, nz=1, dz=0.0):
        points = []
        indexes = []
        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    points.append([dx * i, dy * j, dz * k])
                    indexes.append((i, j, k))
        return np.array(points), np.array(indexes)


class ConvergenceAnalyzer:
    def __init__(self, midpoints, true_variogram):
        self.midpoints = midpoints
        self.true_variogram = true_variogram
        self._deltas = []

    def feed(self, est_variogram):
        self._deltas.append(est_variogram - self.true_variogram)

    @property
    def deltas(self):
        return np.array(self._deltas)



import argparse
import utils
import sysconfig


def parse_arguments():
    parser = argparse.ArgumentParser("find-dependencies")
    parser.add_argument("--use-preprocessor", action="store_true", help="Whether to use the C++ preprocessor on the source files before finding their dependencies")
    parser.add_argument("source_files", nargs="+", help="List of source files to look through")
    parser.add_argument("--include-directories", help="List of directories to include while preprocessing the file(s), seperated by ';'")
    return parser.parse_args()


def run():
    args = parse_arguments()
    include_directories = None
    if args.include_directories:
        include_directories = args.include_directories.split(";")
        include_directories.append(f"{sysconfig.get_path('data')}/include")

    print(' '.join(utils.collect_sources(
        args.source_files,
        use_preprocessor=args.use_preprocessor,
        include_directories=include_directories,
        ignore=include_directories,
    )))


if __name__ == '__main__':
    run()



import gaussianfft.advanced
import numpy as np
from time import perf_counter


class ResourceExtrapolator:
    def __init__(self, nmin=8, nmax=64, nstep=8, ndims=2):
        self.nmin = nmin
        self.nmax = nmax
        self.nstep = nstep
        self.ndims = ndims

        # Constants
        self._grid_length = 1000.0
        self._variogram_range = 500.0

    def extrapolate(self, ex_nmax, vtype):
        data = self.generate(vtype)
        data_array = np.array(data)
        x = data_array[:, 0] ** self.ndims
        y = data_array[:, 1]
        ls_sol = np.linalg.lstsq(x.reshape(x.size, 1), y)
        return x, y, ls_sol[0][0]

    def generate(self, vtype='exponential'):
        vario = gaussianfft.variogram(vtype, self._variogram_range)
        data = []
        for n in range(self.nmin, self.nmax + 1, self.nstep):
            t0 = perf_counter()
            self._simulate(vario, n)
            t1 = perf_counter() - t0
            print('{:4}: {} s'.format(n, t1))
            data.append((n, t1))
        return data

    def _simulate(self, vario, n):
        d = self._grid_length / (n - 1)
        gaussianfft.advanced.simulate(vario,
                                n,
                                d,
                                n if self.ndims > 1 else 1,
                                d,
                                n if self.ndims > 2 else 1,
                                d,
                                padx=n,
                                pady=n,
                                padz=n)


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    re = ResourceExtrapolator(nmin=16, nmax=256, nstep=16, ndims=3)
    en = 512
    x, y, sol = re.extrapolate(en, 'exponential')
    plt.loglog(x, y, 'o')
    lref = 1000
    plt.loglog([lref, en ** re.ndims], [sol * lref, sol * en ** re.ndims])
    plt.title('Line/Square/Cube simulation time')
    plt.legend(['Data', 'Extrapolated'], loc=4)
    plt.xlabel('Number of cells (before padding)')
    plt.ylabel('Time (s)')
    plt.grid()
    plt.show()



from .. import Variogram
from typing import overload

from numpy import ndarray


@overload
def simulate(
        variogram: Variogram,
        nx: int, dx: float,
        ny: int = 1, dy: float = -1.0,
        nz: int = 1, dz: float = -1.0,

        padx: int = -1, pady: int = -1, padz: int = -1,

        sx: float = 1.0, sy: float = 1.0, sz: float = 1.0,
) -> ndarray:
    """
Same as gaussianfft.simulate, but with a few additional advanced and
experimental settings.

Parameters
----------
variogram: gaussianfft.Variogram
    An instance of gaussianfft.Variogram (see gaussianfft.variogram).
nx, ny, nz: int
    Grid size of the simulated field. Only nx is required. Setting ny and/or nz to
    a value less than or equal to 1 reduces the dimension. Default is ny = 1 and
    nz = 1.
dx, dy, dz: float
    Grid resolution in x, y and z directions. dx is always required. dy and dz are
    required if respectively ny and nz are greater than 1.
padx, pady, padz: int
    Grid padding as a number of cells. In gaussianfft.simulate, these are set
    automatically to the values returned by gaussianfft.simulation_size.
sx, sy, sz: float
    Gaussian smoothing parameters to reduce the range. The parameters are the
    values of the smoothing kernel at one variogram range and MUST therefore be
    greater than 0 and less than 1. A value close to or greater than 1 means no
    smoothing.

Returns
-------
out: numpy.ndarray
    See gaussianfft.simulate.
  """
    pass


@overload
def simulate(
        variogram: Variogram,
        nx: int, dx: float,
        ny: int = 1, dy: float = -1.0,
        padx: int = -1, pady: int = -1,
        sx: float = 1.0, sy: float = 1.0,
) -> ndarray:...


@overload
def simulate(
        variogram: Variogram,
        nx: int,
        dx: float,
        padx: int = -1,
        sx: float = 1.0,
) -> ndarray:...



from ._version import __version__
from enum import Enum
from importlib.util import find_spec
if find_spec("numpy") is None:
    raise ImportError("gaussianfft requires NumPy to be installed")


import _gaussianfft

from _gaussianfft import *


class VariogramType(Enum):
    GAUSSIAN = 'gaussian'
    EXPONENTIAL = 'exponential'
    GENERAL_EXPONENTIAL = 'general_exponential'
    SPHERICAL = 'spherical'
    MATERN_32 = 'matern32'
    MATERN_52 = 'matern52'
    MATERN_72 = 'matern72'
    CONSTANT = 'constant'


def variogram(type, *args, **kwargs):
    if isinstance(type, Enum):
        type = type.value
    return _gaussianfft.variogram(type, *args, **kwargs)


__all__ = [
    'variogram', 'simulate', 'seed', 'advanced', 'simulation_size',
    'quote', 'Variogram', 'VariogramType', 'util', 'SizeTVector', 'DoubleVector',
    '__version__',
]



import gaussianfft
import numpy as np
from scipy.sparse.csgraph._traversal import connected_components
from scipy.spatial.distance import cdist, squareform, pdist


class EmpiricalVariogram2:
    reference_point = (0, 0, 0)
    smoothing = 0.00001

    """
    Estimates the variogram, given the provided parameters.
    """
    def __init__(self, v: gaussianfft.Variogram, nx, ny, nz, px, py, pz, close_range, step):
        """
        Creates a variogram estimation class.

        :param v:   A gaussianfft.Variogram instance
        :param nx:  Number of grid cells in x-direction
        :param ny:                          y-direction
        :param nz:                          z-direction
        :param px:  Number of cells to pad with in x-direction
        :param py:                                 y-direction
        :param pz:                                 z-direction
        :param close_range:  Float value defining the range that is considered 'close'. All points within this range are
                             included in the estimation
        :param step:  Sampling resolution in number of cells when beyond close range.
        """
        self.nx, self.ny, self.nz = nx, ny, nz
        self.dx, self.dy, self.dz = 1.0, 1.0, 1.0
        self.px, self.py, self.pz = px, py, pz
        self.v = v
        self.points, self.indexes = EmpiricalVariogram2.find_grid_points(self.nx, self.dx,
                                                                         self.ny, self.dy,
                                                                         self.nz, self.dz)
        self.tuple_indexes = [tuple(i) for i in self.indexes]
        self.dists_from_ref = self._calculate_dists()

        # Indexes to close points, used for estimating the variogram for the short range
        self.ix_short_points = self.dists_from_ref < close_range

        # Indexes for sample points
        self.ix_sample_points = np.array([
            all((j % step == 0) for j in i)
            for i in self.tuple_indexes
        ], dtype=bool)

        self.short_bins = self._find_distance_bins(self.ix_short_points)
        self.sample_bins = self._find_distance_bins(self.ix_sample_points)

        # For validation
        self.s = None

    """ External function """
    def estimate_variogram(self, nmax):
        vdata_short, vdata_sample = self._estimate_variogram_data(nmax)
        x_short = self.short_bins[2]
        x_sample = self.sample_bins[2]
        y_short = np.mean(vdata_short, axis=0)
        y_sample = np.mean(vdata_sample, axis=0)
        y_err_short = np.std(vdata_short, axis=0)
        y_err_sample = np.std(vdata_sample, axis=0)
        return (x_short, y_short, y_err_short), (x_sample, y_sample, y_err_sample)

    """ Initialization """
    def _calculate_dists(self):
        p_idx = self.tuple_indexes.index(EmpiricalVariogram2.reference_point)
        dists = cdist(self.points[p_idx].reshape((1, 3)), self.points)
        return dists.flatten()

    def _find_distance_bins(self, indexes):
        dists = self.dists_from_ref[indexes]
        pairwise_dists = squareform(pdist(dists.reshape((-1, 1))))
        closeness = pairwise_dists < 1e-10
        n_comps, labels = connected_components(closeness, False)
        groups = [
            np.argwhere(labels == i).flatten()
            for i in range(n_comps)
        ]
        group_dists = [dists[g[0]] for g in groups]
        return n_comps, groups, group_dists

    """ Estimation, internal functions """
    def _simulate(self):
        s = gaussianfft.advanced.simulate(self.v,
                                    self.nx, self.dx,
                                    self.ny, self.dy,
                                    self.nz, self.dz,
                                    padx=self.px, pady=self.py, padz=self.pz,
                                    sx=EmpiricalVariogram2.smoothing,
                                    sy=EmpiricalVariogram2.smoothing,
                                    sz=EmpiricalVariogram2.smoothing)
        self.s = np.array(s).reshape((self.nx, self.ny, self.nz), order='F')
        return self.s

    def _estimate_variogram_data(self, nmax):
        short_bin_data = np.zeros((nmax, self.short_bins[0]))
        sample_bin_data = np.zeros((nmax, self.sample_bins[0]))
        for i in range(nmax):
            s = self._simulate()
            d = np.square(self._calculate_all_deltas(s))

            short_d = d[self.ix_short_points]
            sample_d = d[self.ix_sample_points]

            for j, c in enumerate(self.short_bins[1]):
                short_bin_data[i, j] = np.mean(short_d[c])

            for j, c in enumerate(self.sample_bins[1]):
                sample_bin_data[i, j] = np.mean(sample_d[c])

        return short_bin_data, sample_bin_data

    def _calculate_all_deltas(self, grid):
        if self.ny == 1:
            grid = grid.reshape((self.nx, 1, 1))
        elif self.nz == 1:
            grid = grid.reshape((self.nx, self.ny, 1))

        delta = grid - grid[EmpiricalVariogram2.reference_point]
        delta_flat = delta[self.indexes[:, 0], self.indexes[:, 1], self.indexes[:, 2]]
        return delta_flat

    @staticmethod
    def find_grid_points(nx, dx, ny=1, dy=0.0, nz=1, dz=0.0):
        points = []
        indexes = []
        for i in range(nx):
            for j in range(ny):
                for k in range(nz):
                    points.append([dx * i, dy * j, dz * k])
                    indexes.append((i, j, k))
        return np.array(points), np.array(indexes)


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    # Next steps:
    # - structured testing of the parameters. Verify the hypothesis that 4x range is sufficient
    # - 1D verification that the estimation is working as intended.
    nx, ny, nz = 100, 100, 1
    r = 600
    px, py, pz = 300, 300, 0
    v = gaussianfft.variogram('exponential', r, r, r)
    ev = EmpiricalVariogram2(v, nx, ny, nz, px, py, pz, 30.0, 5)
    (x_short, y_short, y_err_short), (x_sample, y_sample, y_err_sample) = ev.estimate_variogram(5000)

    plot_method = 2
    if plot_method == 0:
        # Plot empirical variogram, true variogram and realization

        plt.subplot(121)
        # plt.title('n+p={}, range={}'.format(nx + px, r))
        plt.title('Variogram estimation')
        plt.xlabel('Distance')
        plt.ylabel('Variogram value, $\gamma$')
        plt.plot(x_short, 0.5 * y_short, '.b', label='Empirical')
        plt.plot(x_sample, 0.5 * y_sample, '.b', label='_nolegend_')
        plt.grid()

        x_true = np.linspace(0, np.sqrt(nx ** 2 + ny ** 2 + nz ** 2), 100)
        y_true = [1.0 - v.corr(x) for x in x_true]
        plt.plot(x_true, y_true, '-k', label='True')
        plt.legend(loc=4)
        plt.xlim([0, 141])

        plt.subplot(122)
        plt.imshow(ev.s[:, :, int(ev.s.shape[2]/2)], interpolation='None')
        plt.title('Sample realization')

        plt.show()
    elif plot_method in [1, 2]:
        # Plot empirical variogram, true variogram and difference
        plt.rcParams['figure.figsize'] = [20, 10]
        plt.subplot(121)
        # plt.title('n+p={}, range={}'.format(nx + px, r))
        plt.title('Variogram estimation')
        plt.xlabel('Distance')
        plt.ylabel('Variogram value, $\gamma$')
        plt.plot(x_short, 0.5 * y_short, '.b', label='Empirical')
        plt.plot(x_sample, 0.5 * y_sample, '.b', label='_nolegend_')
        plt.grid()

        x_true = np.linspace(0, np.sqrt(nx ** 2 + ny ** 2 + nz ** 2), 100)
        y_true = [1.0 - v.corr(x) for x in x_true]
        plt.plot(x_true, y_true, '-k', label='True', lw=2)

        plt.xlim([0, 141])

        if plot_method == 2:
            # Include the smoothed variogram
            y_smooth = [1.0 - v.corr(x) * EmpiricalVariogram2.smoothing ** ((x/r)**2) for x in x_true]
            plt.plot(x_true, y_smooth, '-r', lw=2, label='Smoothed')
        plt.legend(loc=4)

        plt.subplot(122)
        diffs = []
        for x, y in zip(x_short, y_short):
            diffs.append(0.5*y - (1.0 - v.corr(x)))
        for x, y in zip(x_sample, y_sample):
            diffs.append(0.5*y - (1.0 - v.corr(x)))
        plt.plot(x_short + x_sample, diffs, '.b')
        plt.xlim([0, 141])
        plt.plot([0, 141], [0, 0], 'k', lw=2)
        plt.grid()
        plt.xlabel('Distance')
        yl = np.max(np.abs(diffs))
        plt.ylim([-yl, yl])
        plt.title('Difference')
        plt.show()




from enum import Enum
from typing import Collection, Iterable, Optional, Union, overload

from numpy import ndarray

import advanced

Number = Union[int, float]

try:
    from typing import Literal

    # More precise type annotations is available in Python 3.8
    VARIOGRAM_TYPE = Literal[
        'gaussian', 'exponential', 'general_exponential',
        'spherical', 'matern32', 'matern52', 'matern72', 'constant'
    ]
except ImportError:
    # Fall back to generic string
    VARIOGRAM_TYPE = str


class VariogramType(Enum):
    GAUSSIAN = 'gaussian'
    EXPONENTIAL = 'exponential'
    GENERAL_EXPONENTIAL = 'general_exponential'
    SPHERICAL = 'spherical'
    MATERN_32 = 'matern32'
    MATERN_52 = 'matern52'
    MATERN_72 = 'matern72'
    CONSTANT = 'constant'


class DoubleVector(object, Collection):
    def append(self, num: Number) -> None: ...

    def extend(self, iterable: Iterable[Number]) -> None: ...


class SizeTVector(object, Collection):
    def append(self, num: Number) -> None: ...

    def extend(self, iterable: Iterable[Number]) -> None: ...


class Variogram(object):
    @overload
    def corr(self, dx: float, dy: float, dz: float) -> float:
        """Correlation function with distance as input for 1D, 2D, and 3D."""
        pass

    @overload
    def corr(self, dx: float, dy: float) -> float:...

    @overload
    def corr(self, dx: float) -> float:...


"""
gaussianfft.variogram
"""


def variogram(
        type: Union[VARIOGRAM_TYPE, VariogramType],
        main_range: float,
        perp_range: Optional[float] = 0.0,
        depth_range: Optional[float] = 0.0,
        azimuth: Optional[float] = 0.0,
        dip: Optional[float] = 0.0,
        power: Optional[float] = 1.5,
) -> Variogram:
    """Factory function for creating a particular variogram. The variogram is always
defined in three directions, but for simulation in fewer dimensions than three,
only the corresponding number of directions are used.

Parameters
----------
type: string
    A string representing the type of variogram. The following types are supported:
    gaussian, exponential, general_exponential, spherical, matern32, matern52,
    matern72 and constant.
main_range: float
    Range of the variogram in the main direction.
perp_range, depth_range: floats, optional
    Parameters representing the range of the variogram in the two directions
    perpendicular to main_range. If any of these are zero, the default is to set
    the value to the same as main_range.
azimuth: float, optional
    Lateral orientation of the variogram in degrees. Default is 0.0.
dip: float, optional
    Dip direction of the variogram in degrees. Default is 0.0
power: float, optional
    Power of the exponent for the general_exponential variogram, which is the only
    variogram type this is used for. Default is 1.5.

Returns
-------
out: Variogram
    An instance of the class gaussianfft.Variogram.

Examples
--------
>>> import gaussianfft as grf
>>> grf.variogram('gaussian', 1000.0)

Specifying dip

>>> grf.variogram('matern52', 1000.0, dip=45.0)

Multiple directions

>>> grf.variogram('general_exponential', 1000.0, 500.0, 250.0, power=1.8)
"""
    pass


"""
gaussianfft.simulate
"""


@overload
def simulate(variogram: Variogram, nx: int, ny: int, nz: int, dx: float, dy: float, dz: float) -> ndarray:
    """
Simulates a Gaussian random field with the corresponding variogram in one, two or
three dimensions. The random generator seed may be set by using gaussianfft.seed.

Parameters
----------
variogram: Variogram
    An instance of gaussianfft.Variogram (see gaussianfft.variogram).
nx, ny, nz: int
    Grid size of the simulated field. Only nx is required. Setting ny and/or nz to
    a value less than or equal to 1 reduces the dimension. Default is ny = 1 and
    nz = 1.
dx, dy, dz: float
    Grid resolution in x, y and z directions. dx is always required. dy and dz are
    required if respectively ny and nz are greater than 1.

Returns
-------
out: numpy.ndarray
    One-dimensional array with the simulation result. Uses Fortran ordering if the
    simulation is multi-dimensional.

Examples
--------
>>> import gaussianfft as grf
>>> v = grf.variogram('gaussian', 250.0, 125.0)
>>> nx, dx = 10, 100.0
>>> z = grf.simulate(v, nx, dx)
>>> z
array([-1.29924289, -1.51172913, -1.2935657 , -0.80779427,  0.22217236,
        1.26740091,  0.66094991, -0.77396656,  0.01523847,  0.44392584])

Multi-dimensional simulation

>>> nx, ny = 100, 200
>>> dx, dy = 10.0, 5.0
>>> z = grf.simulate(v, nx, dx, ny, dy)
>>> z_np = z.reshape((nx, ny), order='F')
>>> z_np.shape
(100,200)
"""
    pass


@overload
def simulate(variogram: Variogram, nx: int, dx: float, ny: int, dy: float, nz: int, dz: float) -> ndarray:...


@overload
def simulate(variogram: Variogram, nx: int, dx: float, ny: int, dy: float) -> ndarray:...


@overload
def simulate(variogram: Variogram, nx: int, dx: float) -> ndarray:...


"""
gaussianfft.seed
"""


@overload
def seed(seed: int) -> None:
    """
Sets the current simulation seed. If this has not been set when calling
gaussianfft.simulate, it is set to the current time with second precision. Be wary of
the latter, in particular if gaussianfft is used in a parallel-processing context.

Examples
--------
>>> import gaussianfft
>>> gaussianfft.seed(123)
    """
    pass


@overload
def seed() -> int:
    """
Gets the current simulation seed. Throws RunTimeError if the seed has not been set
yet.

Examples
--------
>>> import gaussianfft
>>> gaussianfft.seed(123)
>>> gaussianfft.seed()
123
    """
    pass


"""
gaussianfft.simulation_size
"""


@overload
def simulation_size(variogram: Variogram, nx: int, dx: float, ny: int, dy: float, nz: int, dz: float) -> SizeTVector: ...


@overload
def simulation_size(variogram: Variogram, nx: int, dx: float, ny: int, dy: float) -> SizeTVector:
    """
Function for determining the grid size after padding in order to assess the
complexity of the problem. Returns bindings to a vector with up to three elements
with the number of grid cells after the grid has been padded. Signature is the
same as gaussianfft.simulate.

Examples
--------
>>> import gaussianfft as grf

>>> v = grf.variogram('spherical', 250.0, 125.0)
>>> nx, ny, dx, dy = 100, 100, 10.0, 10.0
>>> list(grf.simulation_size(v, nx, dx, ny, dy))
>>>[126, 113]
"""
    pass


@overload
def simulation_size(variogram: Variogram, nx: int, dx: float) -> SizeTVector: ...

"""
gaussianfft.quote
"""

def quote() -> str: ...
